@@ -9,13 +9,15 @@ can work with any of the supported filesystems.
 from __future__ import absolute_import, print_function, unicode_literals
 
 import abc
+import hashlib
 import itertools
 import os
 import threading
 import time
 import typing
 from contextlib import closing
-from functools import partial
+from functools import partial, wraps
+import warnings
 
 import six
 
@@ -61,6 +63,33 @@ if False:  # typing.TYPE_CHECKING
 __all__ = ["FS"]
 
 
+def _new_name(method, old_name):
+    """Return a method with a deprecation warning."""
+    # Looks suspiciously like a decorator, but isn't!
+
+    @wraps(method)
+    def _method(*args, **kwargs):
+        warnings.warn(
+            "method '{}' has been deprecated, please rename to '{}'".format(
+                old_name, method.__name__
+            ),
+            DeprecationWarning,
+        )
+        return method(*args, **kwargs)
+
+    deprecated_msg = """
+        Note:
+            .. deprecated:: 2.2.0
+                Please use `~{}`
+""".format(
+        method.__name__
+    )
+    if getattr(_method, "__doc__"):
+        _method.__doc__ += deprecated_msg
+
+    return _method
+
+
 @six.add_metaclass(abc.ABCMeta)
 class FS(object):
     """Base class for FS objects.
@@ -72,6 +101,9 @@ class FS(object):
     # most FS will use default walking algorithms
     walker_class = Walker
 
+    # default to SubFS, used by opendir and should be returned by makedir(s)
+    subfs_class = None
+
     def __init__(self):
         # type: (...) -> None
         """Create a filesystem. See help(type(self)) for accurate signature.
@@ -80,6 +112,10 @@ class FS(object):
         self._lock = threading.RLock()
         super(FS, self).__init__()
 
+    def __del__(self):
+        """Auto-close the filesystem on exit."""
+        self.close()
+
     def __enter__(self):
         # type: (...) -> FS
         """Allow use of filesystem as a context manager.
@@ -116,12 +152,8 @@ class FS(object):
     # ---------------------------------------------------------------- #
 
     @abc.abstractmethod
-    def getinfo(
-        self,
-        path,  # type: Text
-        namespaces=None,  # type: Optional[Collection[Text]]
-    ):
-        # type: (...) -> Info
+    def getinfo(self, path, namespaces=None):
+        # type: (Text, Optional[Collection[Text]]) -> Info
         """Get information about a resource on a filesystem.
 
         Arguments:
@@ -256,7 +288,7 @@ class FS(object):
         # type: (Text, RawInfo) -> None
         """Set info on a resource.
 
-        This method is the compliment to `~fs.base.FS.getinfo`
+        This method is the complement to `~fs.base.FS.getinfo`
         and is used to set info values on a resource.
 
         Arguments:
@@ -350,7 +382,7 @@ class FS(object):
 
         Example:
             >>> with OSFS('~/Desktop') as desktop_fs:
-            ...    desktop_fs.settext(
+            ...    desktop_fs.writetext(
             ...        'note.txt',
             ...        "Don't forget to tape Game of Thrones"
             ...    )
@@ -383,7 +415,7 @@ class FS(object):
                 raise errors.DestinationExists(dst_path)
             with closing(self.open(src_path, "rb")) as read_file:
                 # FIXME(@althonos): typing complains because open return IO
-                self.setbinfile(dst_path, read_file)  # type: ignore
+                self.upload(dst_path, read_file)  # type: ignore
 
     def copydir(self, src_path, dst_path, create=False):
         # type: (Text, Text, bool) -> None
@@ -393,7 +425,7 @@ class FS(object):
             src_path (str): Path of source directory.
             dst_path (str): Path to destination directory.
             create (bool): If `True`, then ``dst_path`` will be created
-                if it doesn't exist alreadys (defaults to `False`).
+                if it doesn't exist already (defaults to `False`).
 
         Raises:
             fs.errors.ResourceNotFound: If the ``dst_path``
@@ -554,7 +586,7 @@ class FS(object):
             iter_info = itertools.islice(iter_info, start, end)
         return iter_info
 
-    def getbytes(self, path):
+    def readbytes(self, path):
         # type: (Text) -> bytes
         """Get the contents of a file as bytes.
 
@@ -572,7 +604,9 @@ class FS(object):
             contents = read_file.read()
         return contents
 
-    def getfile(self, path, file, chunk_size=None, **options):
+    getbytes = _new_name(readbytes, "getbytes")
+
+    def download(self, path, file, chunk_size=None, **options):
         # type: (Text, BinaryIO, Optional[int], **Any) -> None
         """Copies a file from the filesystem to a file-like object.
 
@@ -595,14 +629,16 @@ class FS(object):
 
         Example:
             >>> with open('starwars.mov', 'wb') as write_file:
-            ...     my_fs.getfile('/movies/starwars.mov', write_file)
+            ...     my_fs.download('/movies/starwars.mov', write_file)
 
         """
         with self._lock:
             with self.openbin(path, **options) as read_file:
                 tools.copy_file_data(read_file, file, chunk_size=chunk_size)
 
-    def gettext(
+    getfile = _new_name(download, "getfile")
+
+    def readtext(
         self,
         path,  # type: Text
         encoding=None,  # type: Optional[Text]
@@ -634,6 +670,8 @@ class FS(object):
             contents = read_file.read()
         return contents
 
+    gettext = _new_name(readtext, "gettext")
+
     def getmeta(self, namespace="standard"):
         # type: (Text) -> Mapping[Text, object]
         """Get meta information regarding a filesystem.
@@ -980,7 +1018,7 @@ class FS(object):
 
     def movedir(self, src_path, dst_path, create=False):
         # type: (Text, Text, bool) -> None
-        """Move contents of directory ``src_path`` to ``dst_path``.
+        """Move directory ``src_path`` to ``dst_path``.
 
         Parameters:
             src_path (str): Path of source directory on the filesystem.
@@ -1029,10 +1067,13 @@ class FS(object):
         with self._lock:
             dir_paths = tools.get_intermediate_dirs(self, path)
             for dir_path in dir_paths:
-                self.makedir(dir_path, permissions=permissions)
-
+                try:
+                    self.makedir(dir_path, permissions=permissions)
+                except errors.DirectoryExists:
+                    if not recreate:
+                        raise
             try:
-                self.makedir(path)
+                self.makedir(path, permissions=permissions)
             except errors.DirectoryExists:
                 if not recreate:
                     raise
@@ -1078,7 +1119,7 @@ class FS(object):
         with self._lock:
             with self.open(src_path, "rb") as read_file:
                 # FIXME(@althonos): typing complains because open return IO
-                self.setbinfile(dst_path, read_file)  # type: ignore
+                self.upload(dst_path, read_file)  # type: ignore
             self.remove(src_path)
 
     def open(
@@ -1149,7 +1190,7 @@ class FS(object):
             factory (callable, optional): A callable that when invoked
                 with an FS instance and ``path`` will return a new FS object
                 representing the sub-directory contents. If no ``factory``
-                is supplied then `~fs.subfs.SubFS` will be used.
+                is supplied then `~fs.subfs_class` will be used.
 
         Returns:
             ~fs.subfs.SubFS: A filesystem representing a sub-directory.
@@ -1161,7 +1202,7 @@ class FS(object):
         """
         from .subfs import SubFS
 
-        _factory = factory or SubFS
+        _factory = factory or self.subfs_class or SubFS
 
         if not self.getbasic(path).is_dir:
             raise errors.DirectoryExpected(path=path)
@@ -1230,7 +1271,7 @@ class FS(object):
             iter_info = itertools.islice(iter_info, start, end)
         return iter_info
 
-    def setbytes(self, path, contents):
+    def writebytes(self, path, contents):
         # type: (Text, bytes) -> None
         # FIXME(@althonos): accept bytearray and memoryview as well ?
         """Copy binary data to a file.
@@ -1248,8 +1289,10 @@ class FS(object):
         with closing(self.open(path, mode="wb")) as write_file:
             write_file.write(contents)
 
-    def setbinfile(self, path, file):
-        # type: (Text, BinaryIO) -> None
+    setbytes = _new_name(writebytes, "setbytes")
+
+    def upload(self, path, file, chunk_size=None, **options):
+        # type: (Text, BinaryIO, Optional[int], **Any) -> None
         """Set a file to the contents of a binary file object.
 
         This method copies bytes from an open binary file to a file on
@@ -1260,21 +1303,28 @@ class FS(object):
             path (str): A path on the filesystem.
             file (io.IOBase): a file object open for reading in
                 binary mode.
+            chunk_size (int, optional): Number of bytes to read at a
+                time, if a simple copy is used, or `None` to use
+                sensible default.
+            **options: Implementation specific options required to open
+                the source file.
 
         Note that the file object ``file`` will *not* be closed by this
         method. Take care to close it after this method completes
         (ideally with a context manager).
 
         Example:
-            >>> with open('myfile.bin') as read_file:
-            ...     my_fs.setbinfile('myfile.bin', read_file)
+            >>> with open('~/movies/starwars.mov', 'rb') as read_file:
+            ...     my_fs.upload('starwars.mov', read_file)
 
         """
         with self._lock:
-            with self.open(path, "wb") as dst_file:
-                tools.copy_file_data(file, dst_file)
+            with self.openbin(path, mode="wb", **options) as dst_file:
+                tools.copy_file_data(file, dst_file, chunk_size=chunk_size)
 
-    def setfile(
+    setbinfile = _new_name(upload, "setbinfile")
+
+    def writefile(
         self,
         path,  # type: Text
         file,  # type: IO
@@ -1294,20 +1344,18 @@ class FS(object):
                 (same as `io.open`).
             newline (str): Newline parameter (same as `io.open`).
 
-        This method will read the contents of a supplied file object,
-        and write to a file on the filesystem. If the destination
-        exists, it will first be truncated.
-
-        If ``encoding`` is supplied, the destination will be opened in
-        text mode.
+        This method is similar to `~FS.upload`, in that it copies data from a
+        file-like object to a resource on the filesystem, but unlike ``upload``,
+        this method also supports creating files in text-mode (if the ``encoding``
+        argument is supplied).
 
         Note that the file object ``file`` will *not* be closed by this
         method. Take care to close it after this method completes
         (ideally with a context manager).
 
         Example:
-            >>> with open('myfile.bin') as read_file:
-            ...     my_fs.setfile('myfile.bin', read_file)
+            >>> with open('myfile.txt') as read_file:
+            ...     my_fs.writefile('myfile.txt', read_file)
 
         """
         mode = "wb" if encoding is None else "wt"
@@ -1318,6 +1366,8 @@ class FS(object):
             ) as dst_file:
                 tools.copy_file_data(file, dst_file)
 
+    setfile = _new_name(writefile, "setfile")
+
     def settimes(self, path, accessed=None, modified=None):
         # type: (Text, Optional[datetime], Optional[datetime]) -> None
         """Set the accessed and modified time on a resource.
@@ -1344,7 +1394,7 @@ class FS(object):
 
         self.setinfo(path, raw_info)
 
-    def settext(
+    def writetext(
         self,
         path,  # type: Text
         contents,  # type: Text
@@ -1377,6 +1427,8 @@ class FS(object):
         ) as write_file:
             write_file.write(contents)
 
+    settext = _new_name(writetext, "settext")
+
     def touch(self, path):
         # type: (Text) -> None
         """Touch a file on the filesystem.
@@ -1444,7 +1496,7 @@ class FS(object):
                 pass
             else:
                 if len(sys_path) > max_sys_path_length:
-                    _msg = "path too long " "(max {max_chars} characters in sys path)"
+                    _msg = "path too long (max {max_chars} characters in sys path)"
                     msg = _msg.format(max_chars=max_sys_path_length)
                     raise errors.InvalidPath(path, msg=msg)
         path = abspath(normpath(path))
@@ -1565,3 +1617,31 @@ class FS(object):
         from .tree import render
 
         render(self, **kwargs)
+
+    def hash(self, path, name):
+        # type: (Text, Text) -> Text
+        """Get the hash of a file's contents.
+
+        Arguments:
+            path(str): A path on the filesystem.
+            name(str): One of the algorithms supported by the hashlib module, e.g. `"md5"`
+
+        Returns:
+            str: The hex digest of the hash.
+
+        Raises:
+            fs.errors.UnsupportedHash: If the requested hash is not supported.
+
+        """
+        _path = self.validatepath(path)
+        try:
+            hash_object = hashlib.new(name)
+        except ValueError:
+            raise errors.UnsupportedHash("hash '{}' is not supported".format(name))
+        with self.openbin(path) as binary_file:
+            while True:
+                chunk = binary_file.read(1024 * 1024)
+                if not chunk:
+                    break
+                hash_object.update(chunk)
+        return hash_object.hexdigest()
